///////////////////////////////////////////////////
//
//   Mental Plugin
//   Quad LFO Mosule based on Batumi
//
//   Strum 2017-19
//   strum@softhome.net
//
///////////////////////////////////////////////////

#include "mental.hpp"


struct LowFrequencyOscillator
{
	float phase = 0.0;
	float pw = 0.5;
	float freq = 1.0;
	bool offset = false;
	bool invert = false;
	dsp::SchmittTrigger resetTrigger;
	LowFrequencyOscillator()
  {
	}
  void setFreq(float freq_to_set)
  {
    freq = freq_to_set;
  }
	void setPitch(float pitch)
  {
		pitch = fminf(pitch, 8.0);
		freq = powf(2.0, pitch);
	}
  void setPhase(float phase_to_set)
  {
    phase = phase_to_set;
  }
  
	void setPulseWidth(float pw_)
  {
		const float pwMin = 0.01;
		pw = clamp(pw_, pwMin, 1.0f - pwMin);
	}
	void setReset(float reset)
  {
		if (resetTrigger.process(reset))
    {
			phase = 0.0;
		}
	}
	void step(float dt)
  {
		float deltaPhase = fminf(freq * dt, 0.5);
		phase += deltaPhase;
		if (phase >= 1.0)
			phase -= 1.0;
	}
	float sin()
  {
		if (offset)
			return 1.0 - cosf(2*M_PI * phase) * (invert ? -1.0 : 1.0);
		else
			return sinf(2*M_PI * phase) * (invert ? -1.0 : 1.0);
	}
	float tri(float x)
  {
		return 4.0 * fabsf(x - roundf(x));
	}
	float tri()
  {
		if (offset)
			return tri(invert ? phase - 0.5 : phase);
		else
			return -1.0 + tri(invert ? phase - 0.25 : phase - 0.75);
	}
	float saw(float x)
  {
		return 2.0 * (x - roundf(x));
	}
	float saw()
  {
		if (offset)
			return invert ? 2.0 * (1.0 - phase) : 2.0 * phase;
		else
			return saw(phase) * (invert ? -1.0 : 1.0);
	}
	float sqr()
  {
		float sqr = (phase < pw) ^ invert ? 1.0 : -1.0;
		return offset ? sqr + 1.0 : sqr;
	}
	float light()
  {
		return sinf(2*M_PI * phase);
	}
};

struct MentalQuadLFO : Module
{
	enum ParamIds
  {
    MODE_BUTTON_PARAM,
		FREQ_PARAM,
		NUM_PARAMS = FREQ_PARAM + 4
	};
	enum InputIds
  {
		FREQ_INPUT,
		RESET_INPUT = FREQ_INPUT + 4,
		NUM_INPUTS = RESET_INPUT + 4
	};
	enum OutputIds
  {
		SIN_OUTPUT,
		TRI_OUTPUT = SIN_OUTPUT + 4,
		SAW_OUTPUT = TRI_OUTPUT + 4,
		SQR_OUTPUT = SAW_OUTPUT + 4,
		NUM_OUTPUTS = SQR_OUTPUT + 4
	};
	enum LightIds
  {
    PHASE_POS_LIGHT,
		PHASE_NEG_LIGHT = PHASE_POS_LIGHT + 4,
		MODE_LIGHTS = PHASE_NEG_LIGHT + 4 ,
		NUM_LIGHTS = MODE_LIGHTS + 5
	};
 
  LowFrequencyOscillator oscillator[4];
  
  dsp::SchmittTrigger mode_button_trigger;
  int mode = 0;
  
	MentalQuadLFO()
  {
		config(NUM_PARAMS, NUM_INPUTS, NUM_OUTPUTS, NUM_LIGHTS);

    configParam(MentalQuadLFO::MODE_BUTTON_PARAM, 0.0, 1.0, 0.0, "");    
  }

	void process(const ProcessArgs& args) override;
  
  json_t *dataToJson() override
  {
		json_t *rootJ = json_object();
    
    // save mode
		json_t *modeJ = json_integer((int)mode);
						
		json_object_set_new(rootJ, "mode", modeJ);    
    return rootJ;
  }
  
  void dataFromJson(json_t *rootJ) override
  {
    // read mode
		json_t *modeJ = json_object_get(rootJ, "mode");
		if (modeJ)
    {
			mode = json_integer_value(modeJ);		
		}  
  }
};

void MentalQuadLFO::process(const ProcessArgs& args)
{
  if (mode_button_trigger.process(params[MODE_BUTTON_PARAM].getValue()))
  {
    mode ++;
    for (int i = 0 ; i < 4 ; i ++)
    {
      oscillator[i].setReset(5.00);
    } 
    if (mode > 4) mode = 0;
    for (int i = 0 ; i < 5 ; i ++)
    {
      lights[MODE_LIGHTS + i].value = 0.0;
    }
  }
  
  lights[MODE_LIGHTS + mode].value = 1.0;
  if (mode == 0)
  {
    for (int i = 0 ; i < 4 ; i++)
    {
      oscillator[i].setPitch((params[FREQ_PARAM + i].getValue() * 10 - 5) + inputs[FREQ_INPUT + i].getVoltage());
      oscillator[i].setReset(inputs[RESET_INPUT + i].getVoltage());
    }
  } else
  if (mode == 1)
  {
    for (int i = 0 ; i < 4 ; i++)
    {
      oscillator[i].setPitch((params[FREQ_PARAM].getValue() * 10 - 5) + inputs[FREQ_INPUT].getVoltage());
      oscillator[i].setReset(inputs[RESET_INPUT].getVoltage());
      oscillator[i].setPhase(oscillator[0].phase + i * 0.25);
    }
  } else
  if (mode == 2)
  {
    for (int i = 0 ; i < 4 ; i++)
    {
      oscillator[i].setPitch((params[FREQ_PARAM].getValue() * 10 - 5) + inputs[FREQ_INPUT].getVoltage());
      oscillator[i].setReset(inputs[RESET_INPUT].getVoltage());
      if (i > 0) oscillator[i].setPhase(oscillator[0].phase + (params[FREQ_PARAM + i].getValue()));
    }
  } else
  if (mode == 3)
  {
    oscillator[0].setPitch((params[FREQ_PARAM].getValue() * 10 - 5) + inputs[FREQ_INPUT].getVoltage());
    oscillator[1].setFreq(oscillator[0].freq / (std::round(params[FREQ_PARAM + 1].getValue() * 11 + 1)));
    oscillator[2].setFreq(oscillator[0].freq / (std::round(params[FREQ_PARAM + 2].getValue() * 11 + 1)));
    oscillator[3].setFreq(oscillator[0].freq / (std::round(params[FREQ_PARAM + 3].getValue() * 11 + 1)));
    oscillator[0].setReset(inputs[RESET_INPUT].getVoltage()); 
    oscillator[1].setReset(inputs[RESET_INPUT].getVoltage());
    oscillator[2].setReset(inputs[RESET_INPUT].getVoltage());
    oscillator[3].setReset(inputs[RESET_INPUT].getVoltage());  
  }
  if (mode == 4)
  {
    oscillator[0].setPitch((params[FREQ_PARAM].getValue() * 10 - 5) + inputs[FREQ_INPUT].getVoltage());
    oscillator[1].setFreq(oscillator[0].freq * (std::round(params[FREQ_PARAM + 1].getValue() * 11 + 1)));
    oscillator[2].setFreq(oscillator[0].freq * (std::round(params[FREQ_PARAM + 2].getValue() * 11 + 1)));
    oscillator[3].setFreq(oscillator[0].freq * (std::round(params[FREQ_PARAM + 3].getValue() * 11 + 1)));
    oscillator[0].setReset(inputs[RESET_INPUT].getVoltage()); 
    oscillator[1].setReset(inputs[RESET_INPUT].getVoltage());
    oscillator[2].setReset(inputs[RESET_INPUT].getVoltage());
    oscillator[3].setReset(inputs[RESET_INPUT].getVoltage());
    if (oscillator[0].phase == 0.0)
    {
      oscillator[1].setReset(5.0);
      oscillator[2].setReset(5.0);
      oscillator[3].setReset(5.0);
    }    
  }
  
  for (int i = 0 ; i < 4 ; i++)
  {
	  
    oscillator[i].step(1.0 / args.sampleRate);

	  outputs[SIN_OUTPUT + i].setVoltage(5.0 * oscillator[i].sin());
	  outputs[TRI_OUTPUT + i].setVoltage(5.0 * oscillator[i].tri());
	  outputs[SAW_OUTPUT + i].setVoltage(5.0 * oscillator[i].saw());
	  outputs[SQR_OUTPUT + i].setVoltage(5.0 * oscillator[i].sqr());

	  //lights[PHASE_POS_LIGHT + i].setBrightnessSmooth(fmaxf(0.0, oscillator[i].light()));
	  //lights[PHASE_NEG_LIGHT + i].setBrightnessSmooth(fmaxf(0.0, -oscillator[i].light()));
    lights[PHASE_NEG_LIGHT + i].setSmoothBrightness(fmaxf(0.0, -oscillator[i].light()), 0.02f);
    lights[PHASE_POS_LIGHT + i].setSmoothBrightness(fmaxf(0.0, oscillator[i].light()), 0.02f);

  } 
}

////////////////////////////////////////////////////////////////////////////////
struct MentalQuadLFOWidget : ModuleWidget
{
	MentalQuadLFOWidget(MentalQuadLFO *module)  
  {
    setModule(module);

    setPanel(APP->window->loadSvg(asset::plugin(pluginInstance, "res/MentalQuadLFO.svg")));

    int x_offset = 10.10;
    for (int i = 0 ; i < 4 ; i++)
    {
  	  addParam(createParam<LEDSliderGreen>(mm2px(Vec(2.792 + i * x_offset, 3.937)), module,MentalQuadLFO::FREQ_PARAM + i));

  	  addInput(createInput<CVInPort>(mm2px(Vec(1.003 + i * x_offset, 61.915)), module, MentalQuadLFO::FREQ_INPUT + i));
  	  addInput(createInput<GateInPort>(mm2px(Vec(1.003 + i * x_offset, 72.858)), module, MentalQuadLFO::RESET_INPUT + i));

  	  addOutput(createOutput<OutPort>(mm2px(Vec(1.003 + i * x_offset, 83.759)), module, MentalQuadLFO::SIN_OUTPUT + i));
  	  addOutput(createOutput<OutPort>(mm2px(Vec(1.003 + i * x_offset, 94.173)), module, MentalQuadLFO::TRI_OUTPUT + i));
  	  addOutput(createOutput<OutPort>(mm2px(Vec(1.003 + i * x_offset, 105.169)), module, MentalQuadLFO::SAW_OUTPUT + i));
  	  addOutput(createOutput<OutPort>(mm2px(Vec(1.003 + i * x_offset, 114.583)), module, MentalQuadLFO::SQR_OUTPUT + i));
    	
      addChild(createLight<SmallLight<GreenRedLight>>(Vec(13 + i * 30, 125), module, MentalQuadLFO::PHASE_POS_LIGHT + i));
    }
    for (int i = 0 ; i < 5 ; i++)
    {
      addChild(createLight<MedLight<BlueLED>>(mm2px(Vec(2.905 + i * 8 , 50.035)), module, MentalQuadLFO::MODE_LIGHTS + i));
    }
    addParam(createParam<LEDButton>(Vec(50, 160), module, MentalQuadLFO::MODE_BUTTON_PARAM));
  }
};

Model *modelMentalQuadLFO = createModel<MentalQuadLFO, MentalQuadLFOWidget>("MentalQuadLFO");